"""
Contains all tests related to the ConfigHandler

NOTE: all tests must be functions with names defined using
    the following format:

    def test_XXXXX(): 
"""

from BeetleETL.Handlers import ETLHandler
from BeetleETL.Handlers import ConfigHandler
import os
import pytest

@pytest.mark.unittest
def test_valid_config_setup(config_path, config_name):
    """ verify that a valid config file opens and is not None """
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    assert ConfigObj != None

@pytest.mark.functest
def test_valid_config_format(config_path, config_name):
    """ verifies that a valid config passes the formatting test """
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    return_value = ConfigObj.verify_format()
    assert return_value == True

@pytest.mark.unittest
def test_verify_format_simple(config_path, config_name):
    """ verifies a simple config is valid """
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    return_value = ConfigObj.verify_format()
    assert ConfigObj.valid == True and return_value == True

@pytest.mark.unittest
def test_verify_format_invalid(config_path, config_name):
    """ verifies an error in the config returns false """
    ConfigObj = ConfigHandler.ConfigHandler(config_path)

    # moddify the config to be invalid
    del ConfigObj.config['connectionInfo']

    # perform verification
    return_value = ConfigObj.verify_format()

    # assert config is not valid
    assert ConfigObj.valid == False and return_value == False

@pytest.mark.unittest
def test_check_key_simple():
    """ test that a value in a dictionary is the correct type """
    test_dict = {"key" : "value"}

    # setup config object
    ConfigObj = ConfigHandler.ConfigHandler()
    
    # assert check_key is fine with a valid key and object
    ConfigObj.check_key("key", test_dict)
    assert ConfigObj.valid == True 

    # assert check_key errors with a invalid key for object
    ConfigObj.check_key("key", {"wrong_key":"value"})
    assert ConfigObj.valid == False


@pytest.mark.unittest
def test_check_key_optional_checks():
    """ test all the optional parameters in check_key for proper function """
    test_dict = {"key" : "value"}

    # setup config object
    ConfigObj = ConfigHandler.ConfigHandler()
    
    # assert check_key passes if expected_type is found
    ConfigObj.check_key("key", test_dict, expected_type=str)
    assert ConfigObj.valid == True 

    # assert check_key fails if expected_type is not found
    ConfigObj.valid = True # reset config status
    ConfigObj.check_key("key", test_dict, expected_type=int)
    assert ConfigObj.valid == False

    # assert check_key passes if key is in valid_values
    ConfigObj.valid = True # reset config status
    ConfigObj.check_key("key", test_dict, valid_values=["value"])
    assert ConfigObj.valid == True

    # assert check_key fails if key is not in valid_values
    ConfigObj.valid = True # reset config status
    ConfigObj.check_key("key", test_dict, valid_values=["wrong_value"])
    assert ConfigObj.valid == False


@pytest.mark.unittest
def test_check_data_field(config_path, config_name):
    """ verify the data field is properly handled """

    ConfigObj = ConfigHandler.ConfigHandler(config_path)

    # ensure no key is named 'data' in config
    if 'data' in ConfigObj.config:
        del ConfigObj.config['data'] 

    # assert absense of 'data' in the config, results in a new dict
    ConfigObj.check_data_field()

    assert 'data' in ConfigObj.config and 'last_mongo_id_pulled' in ConfigObj.config['data']


@pytest.mark.unittest
def test_empty_config_format(config_path, config_name):
    """ verifies an empty dictionary for the config is not valid """
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    ConfigObj.config = {}
    assert ConfigObj.verify_format() == False


""" example of parameterized variables with custom names
@pytest.mark.parametrize("var", ["var_data"], ids=["var_name"])
def test_config(var):
    """ """
    assert 3 == 3
"""




"""
"""
import logging
from daemonize import Daemonize
import lockfile
import signal
import time
import sys


class Daemon():
    """ """
    def __init__(self, ETLHandler, run_interval_hrs=0.1):
        
        self.pid = '/tmp/mongotosql-daemon.pid'
        self.ETL = ETLHandler
        self.run_interval_sec = self.hours_to_sec(run_interval_hrs)

        logging.basicConfig(
            format='%(asctime)s -- %(levelname)s: %(message)s',
            filename='ETLactivity.log',
            level=logging.DEBUG, 
            datefmt='%m/%d/%Y %I:%M:%S %p'
            )

        self.daemon = Daemonize(
            app='MongoToSqlD-Daemon', 
            pid=self.pid, 
            action=self.start_action,
            foreground=True
            )


    def hours_to_sec(self, t):
        """ convert hours to seconds (sec = hrs x 60min x 60sec) """
        return float(t*3600)


    def start_action(self):
        """ runs the program as a linux daemon or windows service
        """
        while True:
            self.ETL.get_from_mongo()
            self.ETL.push_to_sql()
            self.ETL.send_email_update()

            logging.info(" <daemon> pulled from mongo")
            logging.info(" <daemon> sleeping for {} sec".format(self.run_interval_sec))
            time.sleep(self.run_interval_sec)
            

    def start(self):
        """ """
        self.daemon.start()


    def stop(self):
        """ """
        self.daemon.exit()

        


"""
Manages configuration file options
"""

import json
import os
import getpass
import logging
from pydoc import locate
from collections import OrderedDict


class ConfigHandler():
    """ Simple class to manage the config json file
    """

    def __init__(self, config_file=""):
        self.config_path = config_file
        self.config = self.setup_config(config_file)
        self.valid = True # represents the state of the config as usable or not
        
    def setup_config(self, file):
        """ parses config file into

            Arguments: fl {string} -- path of json config file
            Returns: 
                [dict] -- if successful, python dict of config file
                [None] -- if error occurs parsing valid json
        """
        logging.info("Setting up configuration settings based on config file")

        # verify that the json file is a regular file
        # first attempt to open provided file and if fails, looks for default file
        if not os.path.isfile(file):
            logging.error("ERROR: {}: {}".format("FileNotFoundError", "Could not find config"))
            return None     

        # open the json file and return it as a dict
        with open(file, "r") as file_open:
            try:
                json_loaded = json.load(file_open, object_pairs_hook=OrderedDict)
                logging.info("Successfully set up configuration settings")
                
                # check if secure auth should be used for passwords
                try:
                    if json_loaded["useSecureAuthentication"]:
                        logging.info("Using secure authentication")
                        json_loaded = self.setup_auth(json_loaded)
                except Exception as err:
                    logging.error("no parameter for 'useSecureAuthentication', should be true/false\n  -> {}".format(err))
                    return None
                return json_loaded
            except Exception as e:
                logging.error("could not parse config file\n  -> {}".format(e))
                return None

    @staticmethod
    def setup_auth(json_loaded):
        """ prompts the user for mongodb/sql server password 
        """
        json_loaded["connectionInfo"]["mongoServerPass"] = getpass.getpass("Provide Mongodb Password> ")
        json_loaded["connectionInfo"]["sqlServerPass"] = getpass.getpass("Provide Sql Server Password> ")
        if json_loaded["emailInfo"]["useEmailNotify"]:
            json_loaded["emailInfo"]["fromPass"] = getpass.getpass("Provide email Password> ")

        return json_loaded

    def check_key(self, key, obj, 
        expected_type=None, isfile=False, valid_values=None, alt_title=None, verbose=False):
        """ 
            Arguments:
                key {string} -- key to check in obj
                obj {dict} -- obj where key should be
                
            Optional Arguments:
                expected_types {list of type obj} -- the expected typse of obj[key]
                isfile {bool} -- performs additional check if obj[key] is a file
                valid_values {list} -- performs extra check to see if obj[key] is in list
                alt_title {string} -- alternate key to print out
                verbose {string} -- only print out INVALID messages
        """
        if not alt_title:
            alt_title = key
        passes = True
        
        # check if value exists
        value = None
        try:
            value = obj[key]
        except:
            advprint((alt_title, "INVALID", "missing/misspelled"))
            self.valid = False
            return

        # if it exists check type
        if expected_type is not None:
            
            if not isinstance(value, expected_type) and value is not None:
                advprint((
                    alt_title, 
                    "INVALID", 
                    "found: " + type(value).__name__, 
                    "expected: " + expected_type.__name__
                    ))
                passes = False
        
        # check for file if specified
        if isfile:
            if not os.path.isfile(value):
                advprint((alt_title, "INVALID", "not valid path"))
                passes = False
        
        # check if value is in list
        if valid_values is not None and value not in valid_values:
            advprint((
                alt_title, 
                "INVALID", 
                "found: " + value,
                "expecting one: ({})".format(",".join(valid_values))
                ))
            passes = False

        # if the key passes it can be displayed to the user
        if passes:
            if verbose:
                advprint((alt_title, "VALID"))
        else:
            self.valid = False
                 
                
    def verify_format(self):
        """ Checks the config after setup, to make sure a proper format is used

            Returns:
                [bool] -- True if all required fields are valid or False
        """

        # top level keys
        #self.check_key('useSecureAuthentication', self.config, bool)
        if "process" in self.config:
            self.check_key('process', self.config, str, valid_values=("manual", "daemon"))
            if  self.config['process'] != "manual":
                self.check_key('TriggerFrequencyHrs', self.config, int)
       
        # check connectionInfo keys
        if 'connectionInfo' in self.config and isinstance(self.config['connectionInfo'], dict):

            if 'customMongoURI' in self.config['connectionInfo'] and self.config['connectionInfo']['customMongoURI'] != "":
                self.check_key('customMongoURI', self.config['connectionInfo'], str)
            else:
                self.check_key('useMongoURISRV', self.config['connectionInfo'], bool)
                self.check_key('mongoServerHost', self.config['connectionInfo'], str)
                self.check_key('mongoServerUser', self.config['connectionInfo'], str)
                if not self.config['useSecureAuthentication']:
                    self.check_key('mongoServerPass', self.config['connectionInfo'], str)
                if 'useMongoSSL' in self.config['connectionInfo']:
                    self.check_key('useMongoSSL', self.config['connectionInfo'], bool)
            
            self.check_key('mongoDatabase', self.config['connectionInfo'], str)
            self.check_key('mongoCollection', self.config['connectionInfo'], str)            

            self.check_key('useWindowsAuth', self.config['connectionInfo'], bool)
            self.check_key('sqlServerDriver', self.config['connectionInfo'], str)
            self.check_key('sqlServerHost', self.config['connectionInfo'], str)
            self.check_key('sqlServerUser', self.config['connectionInfo'], str)
            self.check_key('sqlServerPass', self.config['connectionInfo'], str)
            self.check_key('sqlServerDatabase', self.config['connectionInfo'], str)
            if not self.config['useSecureAuthentication']:
                    self.check_key('sqlServerPass', self.config['connectionInfo'], str)

        else:
            advprint(('connectionInfo', 'INVALID', 'missing'))
            self.valid = False

        # check sql error handling keys
        if 'sqlErrorHandling' in self.config and isinstance(self.config['sqlErrorHandling'], dict):
            expected = ("break", "skip", "update")
            for key, val in self.config['sqlErrorHandling'].items():
                if val not in expected:
                    location = 'sqlErrorHandling/' + key
                    advprint((location, 'INVALID', 'found: ' + str(val), 'Expected One: ' + str(expected)))
                    self.valid = False
        else:
            advprint(("sqlErrorHandling", 'INVALID', 'missing'))
            self.valid = False

        # check email information keys
        if 'emailInfo' in self.config and isinstance(self.config['emailInfo'], dict):
            self.check_key('useEmailNotify', self.config['emailInfo'], bool)
            if 'useEmailNotify' in self.config['emailInfo'] and self.config['emailInfo']['useEmailNotify'] == True:
                self.check_key('smtpHost', self.config['emailInfo'], str)
                self.check_key('smtpPort', self.config['emailInfo'], int)
                self.check_key('subject', self.config['emailInfo'], str)
                self.check_key('fromEmail', self.config['emailInfo'], str)
                self.check_key('fromPass', self.config['emailInfo'], str)
                self.check_key('toEmail', self.config['emailInfo'], list)
                if len(self.config['emailInfo']['toEmail']) < 1:
                    advprint(('toEmail', 'INVLAID', 'requires at least 1 email'))        
                    self.valid = False
        else:
            advprint(('emailInfo', 'INVALID', 'missing'))
            self.valid = False

        # verify mapping
        if "mapping" in self.config:
            idx = 1
            for m in self.config['mapping']:
                parent = "map"+str(idx)
                if 'sql_dest' in m:
                    self.check_key('schema', m['sql_dest'], str, alt_title=parent+"/schema", verbose=False)
                    self.check_key('db', m['sql_dest'], str, alt_title=parent+"/db", verbose=False)
                    self.check_key('table', m['sql_dest'], str, alt_title=parent+"/table", verbose=False)
                else:
                    advprint((parent+'/sql_dest', "INVALID", 'missing'))
                    self.valid = False

                if 'sql_cols' in m:
                    for val in m['sql_cols'].values():
                        self.check_key('mongo_path', val, str, alt_title=parent+"/mongo_path", verbose=False)
                        if 'target_type' in val:
                            self.check_key('target_type', val, str, 
                                alt_title=parent+"/target_type", 
                                verbose=False,
                                valid_values=("date", "str", "int")
                                )
                        if 'required' in val:
                            self.check_key('required', val, bool, alt_title=parent+"/required")
                        if 'allowMongoUpdate' in val:
                            self.check_key('allowMongoUpdate', val, bool, alt_title=parent+"/allowMongoUpdate")
                            if 'mongoUpdateType' in val:
                                self.check_key('mongoUpdateType', val, str, alt_title=parent+"/mongoUpdateType", valid_values=["str", "int", "float", "date"])
                else:
                    advprint((parent+'/sql_cols', "INVALID", 'missing'))
                    self.valid = False
                idx += 1
        else:
            advprint(('mapping', 'INVALID', 'missing'))
            self.valid = False

        # make sure data field exists and create it if it does not
        self.check_data_field()

        return self.valid

    def check_data_field(self):
        """ verify the data field exists and contains the required fields """

        if 'data' not in self.config:
            self.config['data'] = {}
        
        if 'last_mongo_id_pulled' not in self.config['data']:
            self.config['data']['last_mongo_id_pulled'] = ''

    def save_config(self):
        """ saves the current values int self.config to its path """
        basename = os.path.basename(self.config_path)
        try:
            with open(self.config_path, "w") as file:
                json.dump(self.config, file, indent=4)
                logging.info("Successfully updated config ({}) to file".format(basename))
        except Exception as err:
            logging.critical("Could not write config ({}) to file\n -> {}".format(basename, err))



################### STATIC HELPER FUNCTIONS ###################

def advprint(args):
    """ spaces out elements in a list and prints them """
    s = "{:.<30}".format(args[0])
    s += "".join( "{:<19}".format(i) for i in args[1:] )
    print("Config: " + s) 
    logging.info("Config: " + s)


            



from setuptools import setup, find_packages


setup(
    name='BeetleETL',
    version='1.0.4',
    author='Robby Boney & Jonathon Carothers',
    author_email='robbyb@gointerject.com',
    packages=find_packages(exclude=['contrib', 'docs', 'Tests*']),
    
    # integrates our tests into the package
    # run tests with:
    #  > python setup.py test
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],

    scripts=["BeetleETL/beetle"],
    url='https://gitlab.com/Open-Interject/Beetle-ETL',
    license="Apache License",
    description='Open Source python package for moving data from NoSQL to relational databases.',
    long_description=open('README.md', "r").read(),
    python_requires='>=3',
    install_requires=[
        "pymongo >= 3.6.0",
        "pyodbc >= 4.0.23"
    ],
    classifiers=[]
)


"""
Contains all tests related to the ETLHandler

NOTE: all tests must be functions with names defined using
    the following format:

    def test_XXXXX():
    
"""

from BeetleETL.Handlers import ConfigHandler
from BeetleETL.Handlers import MongoHandler
import logging
import os
import pytest


@pytest.mark.unittest
def test_custom_mongo_uri_overwrites_and_fails(config_path, config_name):
    """ verifies that setting a bad customURI overwrites other settings 
        and returns false when attempting to setup a connection
    """

    # setup config object and moddify the customMongoURI setting
    configObj = ConfigHandler.ConfigHandler(config_path)
    configObj.config['connectionInfo']['customMongoURI'] = ""

    # assert that setting up a connection with a valid uri works
    ValidMongoObj = MongoHandler.MongoHandler(configObj.config)
    assert ValidMongoObj.setup_connection() is True
    
    ValidMongoObj.close_connection() # close connection from the mongo server

    # moddify the custom uri to be a invalid server
    configObj.config['connectionInfo']['customMongoURI'] = "mongo://www.fake.com"

    # setup mongo handler object with config
    InvalidMongoObj = MongoHandler.MongoHandler(configObj.config)

    # assert that setting up a connection fails with a bad custom uri
    assert InvalidMongoObj.setup_connection() is False

    InvalidMongoObj.close_connection() # close connection from the mongo server

@pytest.mark.unittest
def test_setup_connection_pos(config_path, config_name):
    """ positive unit test: verify that valid information connects a returns
        true from setup_connection 
    """

    # setup config object and mongo object
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    MongoObj = MongoHandler.MongoHandler(ConfigObj.config)

    # setup connection and assert its valid
    response = MongoObj.setup_connection()
    assert response == True

    # close connection
    MongoObj.close_connection()

@pytest.mark.unittest
def test_setup_connection_neg(config_path, config_name):
    """ negative unit test: verify that invalid information does not connect and
        returns false from setup_connection 
    """

    # setup config object and mongo object
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    ConfigObj.config['connectionInfo']['mongoServerHost'] = "www.fake.com"
    MongoObj = MongoHandler.MongoHandler(ConfigObj.config)

    # setup connection and assert its valid
    response = MongoObj.setup_connection()
    assert response == False

    # close connection
    MongoObj.close_connection()

@pytest.mark.unittest
def test_get_val_from_dict_simple():
    """ test that a simple recurse finds the correct value"""

    # setup object to use
    test_dict = {"level1":{"level2":{"level3":"value"}}}
    response = []

    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level1", "level2", "level3"],
        out=response
    )

    assert [ "value" ] == response

@pytest.mark.unittest
def test_get_val_from_dict_all():
    """ test that a object with lists will find the correct value """

    # setup object to use
    test_dict = {"level1":[[1,2,3], [4,5,6], [7,8,9]]}
    response = []

    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level1[all][all]"],
        out=response
    )
    
    # assert that the end result matches our expected outcome
    assert [1,2,3,4,5,6,7,8,9] == response

@pytest.mark.unittest
def test_get_val_from_dict_all_nested_obj():
    """ test that a object with lists will find the correct value """

    # setup object to use
    test_dict = {"level1":[[{"ne":"v1"},1], [{"ne":"v2"},4], [{"ne":"v3"},7]]}
    response = []

    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level1[all][all]", "ne"],
        out=response
    )
    
    # assert that the end result matches our expected outcome
    assert ["v1", None, "v2", None, "v3", None] == response

@pytest.mark.unittest
def test_get_val_from_dict_valid_type():
    """ test that a object with lists will find the correct value """

    # setup object to use
    test_dict = {"level1":[[{"ne":"v1"},1], [{"ne":"v2"},4], [{"ne":"v3"},7]]}
    response = []

    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level1[all][all]", "ne"],
        out=response,
        valid_types=["str"]
    )
    
    # assert that the end result matches our expected outcome
    assert ["v1", "v2", "v3"] == response

@pytest.mark.unittest
def test_get_val_from_dict_valid_type_dict():
    """ test that a object with lists will find the correct value """

    # setup object to use
    test_dict = {"level1":[[{"ne":"v1"},1], [{"ne":"v2"},4], [{"ne":"v3"},7]],
        "level2" : {"key1":3, "key2":4}
        }
    response = []

    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level1[all][all]"],
        out=response,
        valid_types=["dict"],
        target_type="str"
    )
    
    # assert that the end result matches our expected outcome
    assert ["{'ne': 'v1'}", "{'ne': 'v2'}", "{'ne': 'v3'}"] == response

    response = []
    # setup instance of mongohandler to use
    MongoHandler.MongoHandler().get_val_from_dict(
        test_dict,
        ["level2"],
        out=response,
        valid_types=["dict"],
        target_type="str"
    )
    assert ["{'key1': 3, 'key2': 4}"] == response or ["{'key2': 4, 'key1': 3}"] == response


@pytest.mark.unittest
def test_cast_to_target_type():
    """ verify that int,str,list and dict are properly handled when casting """

    # check casting int to int
    assert type(MongoHandler.cast_to_target_type(1, "int")) == int

    # check casting int to invalid type returns None
    assert type(MongoHandler.cast_to_target_type(1, "false")) == type(None)

    # check casting dict to int or date fails
    assert type(MongoHandler.cast_to_target_type({"key":"val"}, "int")) == type(None)
    assert type(MongoHandler.cast_to_target_type({"key":"val"}, "date")) == type(None)

    # check casting dict to str passes
    assert type(MongoHandler.cast_to_target_type({"key":"val"}, "str")) == str


@pytest.mark.unittest
def test_pull_required_value_missing(config_path, config_name):
    """ verify that int,str,list and dict are properly handled when casting """

    # setup config object and mongo object
    ConfigObj = ConfigHandler.ConfigHandler(config_path)
    ConfigObj.config['mapping'][0]['sql_cols']['grade'] = {
                                            "mongo_path" : "grades[0].grade",
                                            "required" : True
                                            }  
    MongoObj = MongoHandler.MongoHandler(ConfigObj.config)

    # setup connection and assert its valid
    MongoObj.setup_connection()

    # assert a pull will fail because not all documents will have a value for zipcode
    assert MongoObj.pull_collection() == []

    # close connection
    MongoObj.close_connection()
